using System.Drawing;

namespace View_by_Distance.Shared.Models.Stateless.Methods;

internal abstract class ImageHelper
{

    private static bool StartsWith(byte[] thisBytes, byte[] thatBytes)
    {
        for (int i = 0; i < thatBytes.Length; i += 1)
        {
            if (thisBytes[i] != thatBytes[i])
                return false;
        }
        return true;
    }

    private static short ReadLittleEndianInt16(BinaryReader binaryReader)
    {
        byte[] bytes = new byte[sizeof(short)];
        for (int i = 0; i < sizeof(short); i += 1)
            bytes[sizeof(short) - 1 - i] = binaryReader.ReadByte();
        return BitConverter.ToInt16(bytes, 0);
    }

    private static int ReadLittleEndianInt32(BinaryReader binaryReader)
    {
        byte[] bytes = new byte[sizeof(int)];
        for (int i = 0; i < sizeof(int); i += 1)
            bytes[sizeof(int) - 1 - i] = binaryReader.ReadByte();
        return BitConverter.ToInt32(bytes, 0);
    }

    private static Size DecodeBitmap(BinaryReader binaryReader)
    {
        _ = binaryReader.ReadBytes(16);
        int width = binaryReader.ReadInt32();
        int height = binaryReader.ReadInt32();
        return new Size(width, height);
    }

    private static Size DecodeGif(BinaryReader binaryReader)
    {
        int width = binaryReader.ReadInt16();
        int height = binaryReader.ReadInt16();
        return new Size(width, height);
    }

    private static Size DecodePng(BinaryReader binaryReader)
    {
        _ = binaryReader.ReadBytes(8);
        int width = ReadLittleEndianInt32(binaryReader);
        int height = ReadLittleEndianInt32(binaryReader);
        return new Size(width, height);
    }

    private static Size DecodeJfif(BinaryReader binaryReader)
    {
        while (binaryReader.ReadByte() == 0xff)
        {
            byte marker = binaryReader.ReadByte();
            short chunkLength = ReadLittleEndianInt16(binaryReader);
            if (marker == 0xc0)
            {
                _ = binaryReader.ReadByte();
                int height = ReadLittleEndianInt16(binaryReader);
                int width = ReadLittleEndianInt16(binaryReader);
                return new Size(width, height);
            }
            _ = binaryReader.ReadBytes(chunkLength - 2);
        }
        throw new ArgumentException("Could not recognize image format.");
    }

    /// <summary>
    /// Gets the dimensions of an image.
    /// </summary>
    /// <param name="path">The path of the image to get the dimensions of.</param>
    /// <returns>The dimensions of the specified image.</returns>
    /// <exception cref="ArgumentException">The image was of an unrecognized format.</exception>    
    internal static Size GetDimensions(BinaryReader binaryReader, int? faceRight, int? faceBottom)
    {
        Size? result = null;
        Dictionary<byte[], Func<BinaryReader, Size>> _ImageFormatDecoders = new()
        {
            { new byte[] { 0x42, 0x4D }, DecodeBitmap },
            { new byte[] { 0x47, 0x49, 0x46, 0x38, 0x37, 0x61 }, DecodeGif },
            { new byte[] { 0x47, 0x49, 0x46, 0x38, 0x39, 0x61 }, DecodeGif },
            { new byte[] { 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A }, DecodePng },
            { new byte[] { 0xff, 0xd8 }, DecodeJfif },
        };
        int maxMagicBytesLength = _ImageFormatDecoders.Keys.OrderByDescending(x => x.Length).First().Length;
        byte[] magicBytes = new byte[maxMagicBytesLength];
        for (int i = 0; i < maxMagicBytesLength; i += 1)
        {
            magicBytes[i] = binaryReader.ReadByte();
            foreach (KeyValuePair<byte[], Func<BinaryReader, Size>> kvPair in _ImageFormatDecoders)
            {
                if (!StartsWith(magicBytes, kvPair.Key))
                    continue;
                result = kvPair.Value(binaryReader);
                break;
            }
            if (result is not null)
                break;
        }
        if (result is null)
        {
            if (faceRight is null || faceBottom is null)
                throw new Exception("face is null!");
            result = new(faceRight.Value, faceBottom.Value);
        }
        return result.Value;
    }

    /// <summary>
    /// Gets the dimensions of an image.
    /// </summary>
    /// <param name="path">The path of the image to get the dimensions of.</param>
    /// <returns>The dimensions of the specified image.</returns>
    /// <exception cref="ArgumentException">The image was of an unrecognized format.</exception>
    internal static Size GetDimensions(string path, int? faceRight, int? faceBottom)
    {
        Size result;
        using BinaryReader binaryReader = new(File.OpenRead(path));
        result = GetDimensions(binaryReader, faceRight, faceBottom);
        return result;
    }

}